# -*- coding: utf-8 -*-
"""
/***************************************************************************
 MTOPOpenData
                                 A QGIS plugin
 Acceso a servicios webs del Ministerio de Transporte y Obras Públicas (MTOP) de Uruguay
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2018-06-18
        git sha              : $Format:%H$
        copyright            : (C) 2018 by IDE MTOP
        email                : geoportal@mtop.gub.uy
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from PyQt5.QtCore import QSettings, QTranslator, qVersion, QCoreApplication, QRectF, Qt
from PyQt5.QtGui import QIcon, QPixmap, QTextCursor
from PyQt5.QtWidgets import QAction, QGraphicsScene, QTableWidgetItem, QMessageBox, QHeaderView, QWidget, QGraphicsPixmapItem, QAbstractItemView
from qgis.core import QgsVectorLayer, QgsProject, QgsRasterLayer
# Initialize Qt resources from file resources.py
from .resources import *
# Import the code for the dialog
from .MTOPOpenData_dialog import MTOPOpenDataDialog
import os.path, webbrowser
import sys
import hashlib
import json
import codecs
from .webService_CLASS import WebServiceParams
#from PIL import Image
import struct
import imghdr

if sys.version_info[0] < 3:
    from configparser import SafeConfigParser
else:
    from configparser import SafeConfigParser

class MTOPOpenData:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'MTOPOpenData_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        # Create the dialog (after translation) and keep reference
        self.dlg = MTOPOpenDataDialog()

        # Declare instance attributes
        self.actions = []
        self.menu = 'MTOP Open Data'
        self.language = "ES"
        self.csvWithDatasets = os.path.join(os.path.dirname(__file__),"data.min.json")
        self.quicklooks_dir =  os.path.join(os.path.dirname(__file__),"quicklooks")

        # load config file
        configs = SafeConfigParser()
        configs.read(os.path.join(os.path.dirname(__file__),"MTOPOpenData.cfg"))
        self.sourceUpdates = configs.get('Data', 'sources')
        self.infoUrl = configs.get('Info', 'info')

        # TODO: We are going to let the user set this up in a future iteration
        self.toolbar = self.iface.addToolBar(u'MTOPOpenData')
        self.toolbar.setObjectName(u'MTOPOpenData')

        # define the datasets used. List of webServices custon objects
        self.checkUpdates()
        self.datasets = self.loadDatasets()

        #events
        self.dlg.tableWidget.itemSelectionChanged.connect(self.updateDescAndQL)
        self.dlg.load_btn.released.connect(self.loadWebService)
        self.dlg.close_btn.released.connect(self.close)
        self.init_table()
        self.updateLanguage()
        self.dlg.search_lineEdit.textEdited.connect(self.search)
        self.dlg.info_btn.released.connect(self.showInfo)

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('MTOPOpenData', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/MTOPOpenData/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'MTOPOpenData'),
            callback=self.run,
            parent=self.iface.mainWindow())


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(
                self.tr(u'&MTOPOpenData'),
                action)
            self.iface.removeToolBarIcon(action)
        # remove the toolbar
        del self.toolbar


    def run(self):
        """Run method that performs all the real work"""
        # show the dialog
        self.dlg.show()
        # Run the dialog event loop
        result = self.dlg.exec_()
        # See if OK was pressed
        if result:
            # Do something useful here - delete the line containing pass and
            # substitute with your code.
            pass

    def checkUpdates(self):
        """Check data sources updates"""
        try:
            f = urllib.request.urlopen(self.sourceUpdates)
            data = f.read()
            md5_update = hashlib.md5(data).hexdigest()
            updateFile_dir = os.path.join(os.path.dirname(__file__), "data.min.json.new")
            with open(updateFile_dir, "w+b") as update:
                update.write(data)
                update.close()
        except:
            print("Info: MTOPOpenData: Error checking updates")
            return
        with open(self.csvWithDatasets) as original:
            data = original.read()
            md5_original = hashlib.md5(data).hexdigest()
            original.close()

        if md5_original != md5_update:
            print("Info: MTOPOpenData: New Data Sources Updates!")
            os.remove(self.csvWithDatasets)
            os.rename(updateFile_dir, self.csvWithDatasets)
        else:
            print("Info: MTOPOpenData: No Updates Avalible")
            os.remove(updateFile_dir)

        f.close()

    def loadDatasets(self):
        f = open(self.csvWithDatasets, "rb")
        aux = codecs.getreader("utf-8")
        reader = json.load(aux(f))

        webServicesList = []
        for row in reader:
            #read json data
            nameES = row["Name_ES"]
            nameEn = row["Name_EN"]
            name = nameES
            sourceES = row["Source_ES"]
            sourceEN = row["Source_EN"]
            creationDate = row["Created date"]
            descEN = row["Description_EN"]
            descES = row["Description_ES"]
            layerName = row["Layer_Name"].split(":")
            server = row["server"]
            metadataServer = row["metadata_server"]
            uuid = row["uuid"]

            for service in row["Type"]:
                serviceType = service
                webServiceObj = WebServiceParams(name, nameES,nameEn,sourceES,sourceEN,creationDate,
                                             descEN,descES,serviceType, layerName,server,metadataServer, uuid)
                webServicesList.append(webServiceObj)

        f.close()
        return webServicesList

    @staticmethod
    def get_image_size(fname):    
        with open(fname, 'rb') as fhandle:
            head = fhandle.read(24)
            if len(head) != 24:
                return
            if imghdr.what(fname) == 'png':
                check = struct.unpack('>i', head[4:8])[0]
                if check != 0x0d0a1a0a:
                    return
                width, height = struct.unpack('>ii', head[16:24])
            elif imghdr.what(fname) == 'gif':
                width, height = struct.unpack('<HH', head[6:10])
            elif imghdr.what(fname) == 'jpeg':
                try:
                    fhandle.seek(0)
                    size = 2
                    ftype = 0
                    while not 0xc0 <= ftype <= 0xcf:
                        fhandle.seek(size, 1)
                        byte = fhandle.read(1)
                        while ord(byte) == 0xff:
                            byte = fhandle.read(1)
                        ftype = ord(byte)
                        size = struct.unpack('>H', fhandle.read(2))[0] - 2
                    fhandle.seek(1, 1)
                    height, width = struct.unpack('>HH', fhandle.read(4))
                except Exception:
                    return
            else:
                return
            return width, height

    def updateDescAndQL(self):
        # get the name of the selected dataset
        dataset_name, dataset_serviceType = self.getSelectedNameAndType()

        if not dataset_serviceType or not dataset_name:
            return

        #custom web service object
        dataset = self.selectdataSets(dataset_name, dataset_serviceType)

        # desc = dataset.getDescription(self.language)
        dataset.loadOnlineData(self.language, self.quicklooks_dir)
        # name = dataset.getName(self.language)
        quicklook = os.path.join(self.quicklooks_dir, dataset.QLname)

        #update decription
        self.dlg.textEdit.clear()

        cursor = QTextCursor(self.dlg.textEdit.document())
        cursor.insertHtml("<h3> "+dataset.name+" <br></h3>")
        cursor.insertHtml("<p> "+dataset.desc+" <br><br></p>")
        cursor.insertHtml("<p><i> Ultima actualización del metadato: "+dataset.creationDate+" <br></i></p>")

        self.dlg.textEdit.setReadOnly(True)

        #GET DIMENSIONS OF THE IMAGE
        #img = Image.open(quicklook)
        #w, h = img.size
        w, h = self.get_image_size(quicklook)

        scene = QGraphicsScene()
        pic = QPixmap(quicklook)
        scene.addItem(QGraphicsPixmapItem(pic))

        self.dlg.graphicsView.setScene(scene)
        self.dlg.graphicsView.fitInView(QRectF(0, 0, w, h), Qt.KeepAspectRatio)
        self.dlg.graphicsView.show()

    def loadWebService(self):
        # get the selected row
        dataset_name,dataset_serviceType = self.getSelectedNameAndType()

        #custom web service object
        dataset = self.selectdataSets(dataset_name,dataset_serviceType)

        urlWithParams = dataset.webServiceParams()
        if dataset.serviceType == "WMS":
            rlayer = QgsRasterLayer(dataset.webServiceParams(), dataset.getName(self.language), dataset.serviceType.lower())
            if not rlayer.isValid():
                QMessageBox.information(None, "ERROR:", str(dataset_name + "  cannot be loaded. Check your internet connection."))
            QgsProject.instance().addMapLayer(rlayer)
            #self.iface.legendInterface().setLayerExpanded(rlayer, True) # no collapse the layer'slegend

        elif dataset.serviceType == "WFS":
            vlayer = QgsVectorLayer(dataset.webServiceParams(), dataset.getName(self.language), dataset.serviceType)
            if not vlayer.isValid():
                QMessageBox.information(None, "ERROR:", str(dataset_name + "  cannot be loaded. Check your internet connection."))
            QgsProject.instance().addMapLayer(vlayer)
            #re-appear window
            self.dlg.raise_()
            self.dlg.activateWindow()

    def close(self):
        # prevent update deletes descriptions
        self.dlg.tableWidget.itemSelectionChanged.disconnect(self.updateDescAndQL)
        #clear everything
        self.init_table()
        # clear description and quicklook
        self.dlg.textEdit.clear()
        # clear search box
        #text = self.dlg.search_lineEdit.setText("")
        self.init_searchBox()
        # make a new emptey scene to show
        if not self.dlg.graphicsView is None:
            scene = QGraphicsScene()
            pic = QPixmap()
            scene.addItem(QGraphicsPixmapItem(pic))
            self.dlg.graphicsView.setScene(scene)
            self.dlg.graphicsView.show()
        # restore description updates for next load
        self.dlg.tableWidget.itemSelectionChanged.connect(self.updateDescAndQL)
        """close the dialog"""
        self.dlg.close()

    def init_table(self):
        # fill the table with the entire dataset collection
        self.fill_table(self.datasets)

    def fill_table(self,datasetsList):
        # first delete all elements of table
        self.dlg.tableWidget.setRowCount(0)

        #FILL THE TABLE WIDGET
        # get a sorted version of lirt with WebServiceObjects
        WebServiceObjects = self.sort(datasetsList)

        for dataset in WebServiceObjects:
            index = WebServiceObjects.index(dataset)
            self.dlg.tableWidget.insertRow(index)
            # fill layer name
            self.dlg.tableWidget.setItem(index , 0, QTableWidgetItem(dataset.getName(self.language))) #dataset name
            self.dlg.tableWidget.setItem(index , 1, QTableWidgetItem(dataset.serviceType)) # webservice type
            self.dlg.tableWidget.setItem(index , 2, QTableWidgetItem(dataset.getSource(self.language))) # source organization

        self.setTableWidgetBehavour()

    def sort(self,listOfWebServiceObj):
        """Sorts the datasets by name."""
        # make new list of list[name, obj) and sort it
        sortedList = []
        for WebServiceObj in listOfWebServiceObj:
            name = WebServiceObj.getName(self.language)
            # convert to lowercase
            name = name.lower()
            sortedList.append([name,WebServiceObj])
        #sortedList.sort()
        # build the output list (only with the WebServiceObj)
        outputList = []
        for element in sortedList:
            outputList.append(element[1])

        return outputList

    def setTableWidgetBehavour(self):

        #set rows and columns default sizze and lock it
        self.dlg.tableWidget.setColumnWidth(0,200)
        self.dlg.tableWidget.setColumnWidth(1,45)
        self.dlg.tableWidget.setColumnWidth(2,358)
        self.dlg.tableWidget.horizontalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.dlg.tableWidget.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.dlg.tableWidget.resizeRowsToContents()

        # qtableWidget behavour
        self.dlg.tableWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.dlg.tableWidget.setEditTriggers(QAbstractItemView.NoEditTriggers)

    def updateLanguage(self):
        # get the new language
        language = "Spanish"
        #refill the table
        self.init_table()
        # clear description and quicklook
        self.dlg.textEdit.clear()
        # clear search box
        self.init_searchBox()
        # make a new emptey scene to show
        if not self.dlg.graphicsView is None:
            scene = QGraphicsScene()
            pic = QPixmap()
            scene.addItem(QGraphicsPixmapItem(pic))
            self.dlg.graphicsView.setScene(scene)
            self.dlg.graphicsView.show()

    def init_searchBox(self):
        font = self.dlg.search_lineEdit.font()
        font.setItalic(True)
        self.dlg.search_lineEdit.setFont(font)
        self.dlg.search_lineEdit.setText("")

    def search(self):
        # first set/change the  font of the serach box
        font = self.dlg.search_lineEdit.font()
        font.setItalic(False)
        self.dlg.search_lineEdit.setFont(font)

        # clear description and quicklook
        self.dlg.textEdit.clear()

        # make a new emptey scene to show
        if not self.dlg.graphicsView is None:
            scene = QGraphicsScene()
            pic = QPixmap()
            scene.addItem(QGraphicsPixmapItem(pic))
            self.dlg.graphicsView.setScene(scene)
            self.dlg.graphicsView.show()

        # function the searches for a string in the datasets name, service type and otganization
        text = self.dlg.search_lineEdit.text()
        # convert to lower case and remove greek accents in case of Greek
        text = text.lower()
        foundDatasets = []
        for dataset in self.datasets:
            # use lowercase characters and remove greek accents , to make the comparison
            name = dataset.getName(self.language).lower()
            source = dataset.getSource(self.language).lower()
            serviceType = dataset.serviceType.lower()

            if text in name or text in source or text in serviceType:
                #QMessageBox.information(None, "DEBUG:", str(type(dataset.getName(self.language))))
                foundDatasets.append(dataset)
        #fill the table with the found datasets
        self.fill_table(foundDatasets)

    def showInfo(self):
        #Open Info Main Page
        webbrowser.open_new_tab(self.infoUrl)

    def getSelectedNameAndType(self):
        """Gets the name and service type of the selected dataset."""
        # get the selected row (list with indices)
        selectedIndexes = self.dlg.tableWidget.selectionModel().selectedRows()
        if selectedIndexes:
            row = selectedIndexes[0].row()

            # get name and service typeof selected row
            dataset_name = self.dlg.tableWidget.item(row, 0).text()
            dataset_serviceType = self.dlg.tableWidget.item(row, 1).text()
            return [dataset_name,dataset_serviceType]
        else:
            return [None,None]

    def selectdataSets(self, dataset_name, dataset_serviceType):

        for dataset in self.datasets:
            # get the names both in Green and in English in UTF-8 encoding
            dataset_nameES = dataset.nameES
            dataset_nameEN = dataset.nameEn
            # check which dataset has the name we are looking for
            if dataset_name == dataset_nameES or dataset_name == dataset_nameEN:
                if dataset_serviceType == dataset.serviceType:
                    selectedDataset = dataset
                    break

        return selectedDataset

    